open Ast
open AstPlacement
open Tds
open Tam
open Type
open Code

type t1 = AstPlacement.programme
type t2 = string

let custom_label_prefix = "__custom_label_"

let rec join_with_newlines = function
  | [] -> ""
  | h :: t -> h ^ "\n" ^ join_with_newlines t

let rec addr_of_target = function
  | AstType.Ident info -> (
      match info_ast_to_info info with
      | InfoVar (_, _, offset, _register) -> offset
      | _ -> failwith "cannot get address of non-infovar")
  | AstType.Deref (_, inner) -> addr_of_target inner
  | _ -> failwith "cannot get addr of array yet"

let rec size_of_inner_type = function
  | AstType.Ident info -> (
      match info_ast_to_info info with
      | InfoVar (_, Pointer t, _, _) -> getTaille t
      | InfoVar (_, Array t, _, _) -> getTaille t
      | InfoVar (_, t, _, _) -> getTaille t
      | _ -> failwith "cannot get inner type size")
  | AstType.Deref (_, target) -> size_of_inner_type target
  | AstType.ArrayAccess (Pointer t, _, _) -> getTaille t
  | AstType.ArrayAccess (Array t, _, _) -> getTaille t
  | AstType.ArrayAccess (t, _, _) -> getTaille t

let rec size_of_type = function
  | AstType.Ident info -> (
      match info_ast_to_info info with
      | InfoVar (_, t, _, _) -> getTaille t
      | InfoConst _ -> 1
      | InfoFun (_, t, _) -> getTaille t
      | InfoLabelDef _ -> failwith "cannot get size of label"
      | InfoLabelUse _ -> failwith "cannot get size of label")
  | AstType.Deref (_, t) -> size_of_type t
  | AstType.ArrayAccess (t, _, _) -> getTaille t

let rec tam_of_assignment_target load_value
    (target : AstPlacement.assignement_target) =
  match target with
  | AstType.Deref (_, inner_target) ->
      tam_of_assignment_target load_value inner_target
      ^ loadi (if load_value then size_of_inner_type inner_target else 1)
  | AstType.Ident id -> (
      match info_ast_to_info id with
      | InfoVar (_, id_type, offset, register) ->
          loada offset register
          ^ if load_value then loadi (getTaille id_type) else ""
      | InfoConst (_, value) -> loadl_int value
      | _ ->
          failwith "cannot codegen assignment target for non var and non const")
  | AstType.ArrayAccess (elements_type, target, index_expr) ->
      (* calculer l'indice *)
      tam_of_expression index_expr
      (* calculer le déplacement: indice * taille de l'élément en mémoire *)
      ^ loadl_int (getTaille elements_type)
      ^ subr "IMul"
      (* rajouter à l'indice l'adresse de la base du tableau: on fait un loadi de 1 pour lire ce qui a été calculé par l'analyse de la cible (affectable) *)
      ^ tam_of_assignment_target false target
      ^ loadi 1 ^ subr "IAdd"
      ^
      if load_value then
        (* on charge la valeur qui est à l'adresse qu'on vient de laisser en tête de pile *)
        loadi (getTaille elements_type)
      else ""

and tam_of_expression = function
  | AstType.Booleen v -> loadl_int (if v then 1 else 0)
  | AstType.Entier v -> loadl_int v
  | AstType.Binaire _ as bin -> tam_of_binary bin
  | AstType.Unaire _ as un -> tam_of_unary un
  | AstType.AssignmentTarget target -> tam_of_assignment_target true target
  | AstType.AppelFonction _ as function_call ->
      tam_of_function_call function_call
  | AstType.NewPointer t -> loadl_int (getTaille t) ^ malloc
  | AstType.Address info -> (
      match info_ast_to_info info with
      | InfoVar (_, _, offset, _) -> loada offset "SB"
      | _ -> failwith "codegen: cannot take address of non-infovar")
  | AstType.NewArray (element_type, size) ->
      tam_of_expression size
      ^ loadl_int (getTaille element_type)
      ^ subr "IMul" ^ malloc
  | AstType.ArrayLiteral (elements_type, elements) ->
      let size = List.length elements in
      (* On charge toutes les expressions des éléments du tableau dans l'ordre *)
      String.concat "" (List.map tam_of_expression elements)
      (* On alloue de la place pour un nouveau tableau *)
      ^ loadl_int size
      ^ loadl_int (getTaille elements_type)
      ^ subr "IMul" ^ malloc
      (* On store les éléments dedans, de manière contigüe *)
      ^ storei (getTaille elements_type * size)
      (* MAlloc de 0 renvoie l'adresse de la base du tableau, ce qui permet d'utiliser l'expression dans une déclaration ou assignation *)
      ^ loadl_int 0
      ^ malloc
  | AstType.Null -> loadl_int 0 (* null est l'adresse 0 *)

and tam_of_function_call = function
  | AstType.AppelFonction (info, parameters) -> (
      match info_ast_to_info info with
      | InfoFun (function_name, _return_type, _) ->
          (* Create new scope by pushing all params onto the stack *)
          String.concat " " (List.map tam_of_expression parameters)
          (* run the function code *) ^ call "ST" function_name
      | _ -> failwith "unreachable")
  | _ -> failwith "unreachable"

and tam_of_binary = function
  | AstType.Binaire (op, lhs, rhs) -> (
      tam_of_expression lhs ^ tam_of_expression rhs
      ^
      match op with
      | AstType.Fraction -> call "ST" "norm"
      | AstType.PlusInt -> subr "IAdd"
      | AstType.PlusRat -> call "SB" "radd"
      | AstType.MultInt -> subr "IMul"
      | AstType.MultRat -> call "ST" "rmul"
      | AstType.EquInt -> subr "IEq"
      | AstType.EquBool -> subr "IEq"
      | AstType.Inf -> subr "ILss")
  | _ -> failwith "unreachable"

and tam_of_unary = function
  | AstType.Unaire (operator, operand) -> (
      tam_of_expression operand
      ^
      match operator with
      | AstType.Denominateur -> pop 1 1
      | AstType.Numerateur -> pop 0 1)
  | _ -> failwith "unreachable"

let rec tam_of_assignment = function
  | AstPlacement.Affectation (target, value) ->
      tam_of_expression value
      ^ tam_of_assignment_target false target
      ^ storei (size_of_type target)
  | _ -> failwith "wrong codegen called"

and tam_of_declaration = function
  | AstPlacement.Declaration (target, value) -> (
      match info_ast_to_info target with
      | InfoVar (_, typ, offset, register) ->
          push (getTaille typ)
          ^ tam_of_expression value
          ^ store (getTaille typ) offset register
      | _ -> failwith "")
  | _ -> failwith "unreachable"

and tam_of_function (AstPlacement.Fonction (info, _params_info, body)) =
  match info_ast_to_info info with
  | InfoFun (name, _return_type, _params_types) ->
      label name ^ tam_of_block body ^ (* prevent stack overflows *) halt
  | _ -> failwith ""

and tam_of_goto = function 
| (AstPlacement.Goto (label_info)) -> (
  match info_ast_to_info label_info with
  | InfoLabelDef (label_name) -> jump (custom_label_prefix ^ label_name)
  | _ -> failwith "cannot goto non-label"
)
| _ -> failwith "unreachable"

and tam_of_block (block : AstPlacement.bloc) =
  let dispatch_statement_converter = function
    | Affectation _ -> tam_of_assignment
    | Conditionnelle _ -> tam_of_if_then_else
    | TantQue _ -> tam_of_while_loop
    | Retour _ -> tam_of_return
    | AffichageBool _ -> tam_of_print_bool
    | AffichageInt _ -> tam_of_print_int
    | AffichageRat _ -> tam_of_print_rat
    | AffichageAddresse _ -> tam_of_print_addr
    | Declaration _ -> tam_of_declaration
    | For _ -> tam_of_for_loop
    | Goto _ -> tam_of_goto 
    | Empty -> fun _ -> ""
    | Labelled _ -> failwith "dispatch of labelled instruction should be done beforehand"
  in
  let statements, size = block in
  join_with_newlines
    (List.map
       (function 
       (* préfixe __custom_label pour éviter les conflits avec les labels internes *)
       | Labelled(label_info, inner_statement) -> (match info_ast_to_info label_info with
       | InfoLabelDef(label_name) ->
       label (custom_label_prefix ^ label_name) ^ dispatch_statement_converter inner_statement inner_statement
       | _ -> failwith "cannot label instruction with non-infolabel")
       | statement -> dispatch_statement_converter statement statement)
       statements)
  ^ pop 0 size

and tam_of_if_then_else = function
  | AstPlacement.Conditionnelle (condition, then_block, else_block) ->
      (*Push one to store the boolean condition *)
      let else_label = getEtiquette () in
      let end_label = getEtiquette () in
      push (getTaille Bool)
      ^ tam_of_expression condition
      (* jump to else block if condition false *) ^ jumpif 0 else_label
      (* if block *) ^ tam_of_block then_block
      ^ jump end_label (* else block marker *) ^ label else_label
      ^ tam_of_block else_block ^ label end_label
  | _ -> failwith "unreachable"

and tam_of_while_loop = function
  | AstPlacement.TantQue (stop_condition, body) ->
      let loop_start = getEtiquette () in
      let loop_end = getEtiquette () in
      label loop_start
      ^ push (getTaille Bool)
      ^ tam_of_expression stop_condition
      ^ jumpif 0 loop_end ^ tam_of_block body ^ jump loop_start ^ label loop_end
  | _ -> failwith "unreachable"

and tam_of_for_loop = function
  | AstPlacement.For (init, test, incr, body) ->
      let loop_start = getEtiquette () in
      let loop_end = getEtiquette () in
      tam_of_declaration init ^ label loop_start ^ tam_of_expression test
      ^ jumpif 0 loop_end ^ tam_of_block body ^ tam_of_assignment incr
      ^ jump loop_start ^ label loop_end
  | _ -> failwith "unreachable"

and tam_of_return = function
  | AstPlacement.Retour (expr, return_type_size, params_size) ->
      tam_of_expression expr ^ return return_type_size params_size
  | _ -> failwith "unreachable"

and tam_of_print_bool = function
  | AstPlacement.AffichageBool expr ->
      tam_of_expression expr ^ subr "Bout" (* ^ loadl_char '\n' ^ subr "COut" *)
  | _ -> failwith "unreachable"

and tam_of_print_int = function
  | AstPlacement.AffichageInt expr ->
      tam_of_expression expr ^ subr "IOut" (* ^ loadl_char '\n' ^ subr "COut" *)
  | _ -> failwith "unreachable"

and tam_of_print_addr = function
  | AstPlacement.AffichageAddresse target ->
      loadl_int (addr_of_target target)
      ^ subr "IOut" (* ^ loadl_char '\n' ^ subr "COut" *)
  | _ -> failwith "unreachable"

and tam_of_print_rat = function
  | AstPlacement.AffichageRat expr ->
      tam_of_expression expr
      ^ call "SB" "Rout" (* ^ loadl_char '\n' ^ subr "COut" *)
  | _ -> failwith "unreachable"

let analyser (AstPlacement.Programme (functions, main_block)) =
  getEntete ()
  ^ List.fold_left ( ^ ) "\n" (List.map tam_of_function functions)
  ^ label "main" ^ tam_of_block main_block ^ halt
